@@ -21,6 +21,16 @@ void grid_resize_and_reflow(
     size_t tracking_points_count,
     struct coord *const _tracking_points[static tracking_points_count]);
 
+/* Convert row numbers between scrollback-relative and absolute coordinates */
+int grid_row_abs_to_sb(const struct grid *grid, int screen_rows, int abs_row);
+int grid_row_sb_to_abs(const struct grid *grid, int screen_rows, int sb_rel_row);
+
+int grid_sb_start_ignore_uninitialized(const struct grid *grid, int screen_rows);
+int grid_row_abs_to_sb_precalc_sb_start(
+    const struct grid *grid, int sb_start, int abs_row);
+int grid_row_sb_to_abs_precalc_sb_start(
+    const struct grid *grid, int sb_start, int sb_rel_row);
+
 static inline int
 grid_row_absolute(const struct grid *grid, int row_no)
 {
