@@ -81,9 +81,11 @@ pipe_closed:
     return true;
 }
 
+static void alternate_scroll(struct seat *seat, int amount, int button);
+
 static bool
 execute_binding(struct seat *seat, struct terminal *term,
-                const struct key_binding *binding, uint32_t serial)
+                const struct key_binding *binding, uint32_t serial, int amount)
 {
     const enum bind_action_normal action = binding->action;
 
@@ -115,6 +117,14 @@ execute_binding(struct seat *seat, struct terminal *term,
         }
         break;
 
+    case BIND_ACTION_SCROLLBACK_UP_MOUSE:
+        if (term->grid == &term->alt) {
+            if (term->alt_scrolling)
+                alternate_scroll(seat, amount, BTN_BACK);
+        } else
+                cmd_scrollback_up(term, amount);
+        break;
+
     case BIND_ACTION_SCROLLBACK_DOWN_PAGE:
         if (term->grid == &term->normal) {
             cmd_scrollback_down(term, term->rows);
@@ -136,6 +146,14 @@ execute_binding(struct seat *seat, struct terminal *term,
         }
         break;
 
+    case BIND_ACTION_SCROLLBACK_DOWN_MOUSE:
+        if (term->grid == &term->alt) {
+            if (term->alt_scrolling)
+                alternate_scroll(seat, amount, BTN_FORWARD);
+        } else
+            cmd_scrollback_down(term, amount);
+        break;
+
     case BIND_ACTION_SCROLLBACK_HOME:
         if (term->grid == &term->normal) {
             cmd_scrollback_up(term, term->grid->num_rows);
@@ -454,6 +472,11 @@ execute_binding(struct seat *seat, struct terminal *term,
             term, seat->mouse.col, seat->mouse.row, SELECTION_WORD_WISE, true);
         return true;
 
+    case BIND_ACTION_SELECT_QUOTE:
+        selection_start(
+            term, seat->mouse.col, seat->mouse.row, SELECTION_QUOTE_WISE, false);
+        break;
+
     case BIND_ACTION_SELECT_ROW:
         selection_start(
             term, seat->mouse.col, seat->mouse.row, SELECTION_LINE_WISE, false);
@@ -1478,7 +1501,7 @@ key_press_release(struct seat *seat, struct terminal *term, uint32_t serial,
             /* Match translated symbol */
             if (bind->k.sym == sym &&
                 bind->mods == (bind_mods & ~bind_consumed) &&
-                execute_binding(seat, term, bind, serial))
+                execute_binding(seat, term, bind, serial, 1))
             {
                 goto maybe_repeat;
             }
@@ -1489,7 +1512,7 @@ key_press_release(struct seat *seat, struct terminal *term, uint32_t serial,
             /* Match untranslated symbols */
             for (size_t i = 0; i < raw_count; i++) {
                 if (bind->k.sym == raw_syms[i] &&
-                    execute_binding(seat, term, bind, serial))
+                    execute_binding(seat, term, bind, serial, 1))
                 {
                     goto maybe_repeat;
                 }
@@ -1498,7 +1521,7 @@ key_press_release(struct seat *seat, struct terminal *term, uint32_t serial,
             /* Match raw key code */
             tll_foreach(bind->k.key_codes, code) {
                 if (code->item == key &&
-                    execute_binding(seat, term, bind, serial))
+                    execute_binding(seat, term, bind, serial, 1))
                 {
                     goto maybe_repeat;
                 }
@@ -1751,6 +1774,28 @@ mouse_coord_pixel_to_cell(struct seat *seat, const struct terminal *term,
         seat->mouse.row = (y - term->margins.top) / term->cell_height;
 }
 
+static bool
+touch_is_active(const struct seat *seat)
+{
+    if (seat->wl_touch == NULL) {
+        return false;
+    }
+
+    switch (seat->touch.state) {
+    case TOUCH_STATE_IDLE:
+    case TOUCH_STATE_INHIBITED:
+        return false;
+
+    case TOUCH_STATE_HELD:
+    case TOUCH_STATE_DRAGGING:
+    case TOUCH_STATE_SCROLLING:
+        return true;
+    }
+
+    BUG("Bad touch state: %d", seat->touch.state);
+    return false;
+}
+
 static void
 wl_pointer_enter(void *data, struct wl_pointer *wl_pointer,
                  uint32_t serial, struct wl_surface *surface,
@@ -1764,26 +1809,15 @@ wl_pointer_enter(void *data, struct wl_pointer *wl_pointer,
 
     struct seat *seat = data;
 
-    if (seat->wl_touch != NULL) {
-        switch (seat->touch.state) {
-        case TOUCH_STATE_IDLE:
-            mouse_button_state_reset(seat);
-            seat->touch.state = TOUCH_STATE_INHIBITED;
-            break;
-
-        case TOUCH_STATE_INHIBITED:
-            break;
-
-        case TOUCH_STATE_HELD:
-        case TOUCH_STATE_DRAGGING:
-        case TOUCH_STATE_SCROLLING:
-            return;
-        }
-    }
-
     struct wl_window *win = wl_surface_get_user_data(surface);
     struct terminal *term = win->term;
 
+    seat->mouse_focus = term;
+    term->active_surface = term_surface_kind(term, surface);
+
+    if (touch_is_active(seat))
+        return;
+
     int x = wl_fixed_to_int(surface_x) * term->scale;
     int y = wl_fixed_to_int(surface_y) * term->scale;
 
@@ -1799,9 +1833,6 @@ wl_pointer_enter(void *data, struct wl_pointer *wl_pointer,
 
     xassert(tll_length(seat->mouse.buttons) == 0);
 
-    seat->mouse_focus = term;
-    term->active_surface = term_surface_kind(term, surface);
-
     wayl_reload_xcursor_theme(seat, term->scale); /* Scale may have changed */
     term_xcursor_update_for_seat(term, seat);
 
@@ -1837,10 +1868,19 @@ wl_pointer_leave(void *data, struct wl_pointer *wl_pointer,
     struct seat *seat = data;
 
     if (seat->wl_touch != NULL) {
-        if (seat->touch.state != TOUCH_STATE_INHIBITED) {
+        switch (seat->touch.state) {
+        case TOUCH_STATE_IDLE:
+            break;
+
+        case TOUCH_STATE_INHIBITED:
+            seat->touch.state = TOUCH_STATE_IDLE;
+            break;
+
+        case TOUCH_STATE_HELD:
+        case TOUCH_STATE_DRAGGING:
+        case TOUCH_STATE_SCROLLING:
             return;
         }
-        seat->touch.state = TOUCH_STATE_IDLE;
     }
 
     struct terminal *old_moused = seat->mouse_focus;
@@ -1919,7 +1959,7 @@ wl_pointer_motion(void *data, struct wl_pointer *wl_pointer,
     struct seat *seat = data;
 
     /* Touch-emulated pointer events have wl_pointer == NULL. */
-    if (wl_pointer != NULL && seat->touch.state != TOUCH_STATE_INHIBITED)
+    if (wl_pointer != NULL && touch_is_active(seat))
         return;
 
     struct wayland *wayl = seat->wayl;
@@ -2135,6 +2175,95 @@ fdm_csd_move(struct fdm *fdm, int fd, int events, void *data)
     return true;
 }
 
+static const struct key_binding *
+ match_mouse_binding(const struct seat *seat, const struct terminal *term,
+                     int button)
+{
+    if (seat->wl_keyboard != NULL && seat->kbd.xkb_state != NULL) {
+        /* Seat has keyboard - use mouse bindings *with* modifiers */
+
+        const struct key_binding_set *bindings =
+            key_binding_for(term->wl->key_binding_manager, term->conf, seat);
+        xassert(bindings != NULL);
+
+        xkb_mod_mask_t mods;
+        get_current_modifiers(seat, &mods, NULL, 0);
+        mods &= seat->kbd.bind_significant;
+
+        /* Ignore selection override modifiers when
+         * matching modifiers */
+        mods &= ~bindings->selection_overrides;
+
+        const struct key_binding *match = NULL;
+
+        tll_foreach(bindings->mouse, it) {
+            const struct key_binding *binding = &it->item;
+
+            if (binding->m.button != button) {
+                /* Wrong button */
+                continue;
+            }
+
+            if (binding->mods != mods) {
+                /* Modifier mismatch */
+                continue;
+            }
+
+            if (binding->m.count > seat->mouse.count) {
+                /* Not correct click count */
+                continue;
+            }
+
+            if (match == NULL || binding->m.count > match->m.count)
+                match = binding;
+        }
+
+        return match;
+    }
+
+    else {
+        /* Seat does NOT have a keyboard - use mouse bindings *without*
+         * modifiers */
+        const struct config_key_binding *match = NULL;
+        const struct config *conf = term->conf;
+
+        for (size_t i = 0; i < conf->bindings.mouse.count; i++) {
+            const struct config_key_binding *binding =
+                &conf->bindings.mouse.arr[i];
+
+            if (binding->m.button != button) {
+                /* Wrong button */
+                continue;
+            }
+
+            if (binding->m.count > seat->mouse.count) {
+                /* Incorrect click count */
+                continue;
+            }
+
+            const struct config_key_modifiers no_mods = {0};
+            if (memcmp(&binding->modifiers, &no_mods, sizeof(no_mods)) != 0) {
+                /* Binding has modifiers */
+                continue;
+            }
+
+            if (match == NULL || binding->m.count > match->m.count)
+                match = binding;
+        }
+
+        if (match != NULL) {
+            static struct key_binding bind;
+            bind.action = match->action;
+            bind.aux = &match->aux;
+            return &bind;
+        }
+
+        return NULL;
+    }
+
+    BUG("should not get here");
+}
+
 static void
 wl_pointer_button(void *data, struct wl_pointer *wl_pointer,
                   uint32_t serial, uint32_t time, uint32_t button, uint32_t state)
@@ -2147,7 +2276,7 @@ wl_pointer_button(void *data, struct wl_pointer *wl_pointer,
     struct seat *seat = data;
 
     /* Touch-emulated pointer events have wl_pointer == NULL. */
-    if (wl_pointer != NULL && seat->touch.state != TOUCH_STATE_INHIBITED)
+    if (wl_pointer != NULL && touch_is_active(seat))
         return;
 
     struct wayland *wayl = seat->wayl;
@@ -2162,6 +2291,10 @@ wl_pointer_button(void *data, struct wl_pointer *wl_pointer,
     bool send_to_client = false;
 
     if (state == WL_POINTER_BUTTON_STATE_PRESSED) {
+        if (seat->wl_touch != NULL && seat->touch.state == TOUCH_STATE_IDLE) {
+            seat->touch.state = TOUCH_STATE_INHIBITED;
+        }
+
         /* Time since last click */
         struct timespec now, since_last;
         clock_gettime(CLOCK_MONOTONIC, &now);
@@ -2263,6 +2396,12 @@ wl_pointer_button(void *data, struct wl_pointer *wl_pointer,
             }
         }
 
+        if (seat->wl_touch != NULL && seat->touch.state == TOUCH_STATE_INHIBITED) {
+            if (tll_length(seat->mouse.buttons) == 0) {
+                seat->touch.state = TOUCH_STATE_IDLE;
+            }
+        }
+
         if (!have_button) {
             /*
              * Seen on Sway with slurp
@@ -2399,86 +2538,11 @@ wl_pointer_button(void *data, struct wl_pointer *wl_pointer,
             bool consumed = false;
 
             if (cursor_is_on_grid && term_mouse_grabbed(term, seat)) {
-                if (seat->wl_keyboard != NULL && seat->kbd.xkb_state != NULL) {
-                    /* Seat has keyboard - use mouse bindings *with* modifiers */
+                const struct key_binding *match =
+                    match_mouse_binding(seat, term, button);
 
-                    const struct key_binding_set *bindings = key_binding_for(
-                        wayl->key_binding_manager, term->conf, seat);
-                    xassert(bindings != NULL);
-
-                    xkb_mod_mask_t mods;
-                    get_current_modifiers(seat, &mods, NULL, 0);
-                    mods &= seat->kbd.bind_significant;
-
-                    /* Ignore selection override modifiers when
-                     * matching modifiers */
-                    mods &= ~bindings->selection_overrides;
-
-                    const struct key_binding *match = NULL;
-
-                    tll_foreach(bindings->mouse, it) {
-                        const struct key_binding *binding = &it->item;
-
-                        if (binding->m.button != button) {
-                            /* Wrong button */
-                            continue;
-                        }
-
-                        if (binding->mods != mods) {
-                            /* Modifier mismatch */
-                            continue;
-                        }
-
-                        if  (binding->m.count > seat->mouse.count) {
-                            /* Not correct click count */
-                            continue;
-                        }
-
-                        if (match == NULL || binding->m.count > match->m.count)
-                            match = binding;
-                    }
-
-                    if (match != NULL)
-                        consumed = execute_binding(seat, term, match, serial);
-                }
-
-                else {
-                    /* Seat does NOT have a keyboard - use mouse bindings *without* modifiers */
-                    const struct config_key_binding *match = NULL;
-                    const struct config *conf = term->conf;
-
-                    for (size_t i = 0; i < conf->bindings.mouse.count; i++) {
-                        const struct config_key_binding *binding =
-                            &conf->bindings.mouse.arr[i];
-
-                        if (binding->m.button != button) {
-                            /* Wrong button */
-                            continue;
-                        }
-
-                        if (binding->m.count > seat->mouse.count) {
-                            /* Incorrect click count */
-                            continue;
-                        }
-
-                        const struct config_key_modifiers no_mods = {0};
-                        if (memcmp(&binding->modifiers, &no_mods, sizeof(no_mods)) != 0) {
-                            /* Binding has modifiers */
-                            continue;
-                        }
-
-                        if (match == NULL || binding->m.count > match->m.count)
-                            match = binding;
-                    }
-
-                    if (match != NULL) {
-                        struct key_binding bind = {
-                            .action = match->action,
-                            .aux = &match->aux,
-                        };
-                        consumed = execute_binding(seat, term, &bind, serial);
-                    }
-                }
+                if (match != NULL)
+                    consumed = execute_binding(seat, term, match, serial, 1);
             }
 
             send_to_client = !consumed && cursor_is_on_grid;
@@ -2553,26 +2617,15 @@ mouse_scroll(struct seat *seat, int amount, enum wl_pointer_axis axis)
     amount = abs(amount);
 
     if (term_mouse_grabbed(term, seat)) {
-        if (term->grid == &term->alt) {
-            if (term->alt_scrolling) {
-                switch (button) {
-                case BTN_BACK:
-                case BTN_FORWARD:
-                    alternate_scroll(seat, amount, button);
-                    break;
-                }
-            }
-        } else {
-            switch (button) {
-            case BTN_BACK:
-                cmd_scrollback_up(term, amount);
-                break;
+        seat->mouse.count = 1;
 
-            case BTN_FORWARD:
-                cmd_scrollback_down(term, amount);
-                break;
-            }
-        }
+        const struct key_binding *match =
+            match_mouse_binding(seat, term, button);
+
+        if (match != NULL)
+            execute_binding(seat, term, match, seat->pointer.serial, amount);
+
+        seat->mouse.last_released_button = button;
     }
 
     else if (seat->mouse.col >= 0 && seat->mouse.row >= 0) {
@@ -2610,7 +2663,7 @@ wl_pointer_axis(void *data, struct wl_pointer *wl_pointer,
 {
     struct seat *seat = data;
 
-    if (seat->touch.state != TOUCH_STATE_INHIBITED)
+    if (touch_is_active(seat))
         return;
 
     if (seat->mouse.have_discrete)
@@ -2643,7 +2696,7 @@ wl_pointer_axis_discrete(void *data, struct wl_pointer *wl_pointer,
 {
     struct seat *seat = data;
 
-    if (seat->touch.state != TOUCH_STATE_INHIBITED)
+    if (touch_is_active(seat))
         return;
 
     seat->mouse.have_discrete = true;
@@ -2663,7 +2716,7 @@ wl_pointer_frame(void *data, struct wl_pointer *wl_pointer)
 {
     struct seat *seat = data;
 
-    if (seat->touch.state != TOUCH_STATE_INHIBITED)
+    if (touch_is_active(seat))
         return;
 
     seat->mouse.have_discrete = false;
@@ -2681,7 +2734,7 @@ wl_pointer_axis_stop(void *data, struct wl_pointer *wl_pointer,
 {
     struct seat *seat = data;
 
-    if (seat->touch.state != TOUCH_STATE_INHIBITED)
+    if (touch_is_active(seat))
         return;
 
     xassert(axis < ALEN(seat->mouse.aggregated));
@@ -2738,8 +2791,6 @@ wl_touch_down(void *data, struct wl_touch *wl_touch, uint32_t serial,
     struct wl_window *win = wl_surface_get_user_data(surface);
     struct terminal *term = win->term;
 
-    term->active_surface = term_surface_kind(term, surface);
-
     LOG_DBG("touch_down: touch=%p, x=%d, y=%d", (void *)wl_touch,
             wl_fixed_to_int(surface_x), wl_fixed_to_int(surface_y));
 
@@ -2754,6 +2805,7 @@ wl_touch_down(void *data, struct wl_touch *wl_touch, uint32_t serial,
     seat->touch.serial = serial;
     seat->touch.time = time + term->conf->touch.long_press_delay;
     seat->touch.surface = surface;
+    seat->touch.surface_kind = term_surface_kind(term, surface);
     seat->touch.id = id;
 }
 
@@ -2771,7 +2823,10 @@ wl_touch_up(void *data, struct wl_touch *wl_touch, uint32_t serial,
     struct wl_window *win = wl_surface_get_user_data(seat->touch.surface);
     struct terminal *term = win->term;
 
+    struct terminal *old_term = seat->mouse_focus;
+    enum term_surface old_active_surface = term->active_surface;
     seat->mouse_focus = term;
+    term->active_surface = seat->touch.surface_kind;
 
     switch (seat->touch.state) {
     case TOUCH_STATE_HELD:
@@ -2793,7 +2848,8 @@ wl_touch_up(void *data, struct wl_touch *wl_touch, uint32_t serial,
         break;
     }
 
-    seat->mouse_focus = NULL;
+    seat->mouse_focus = old_term;
+    term->active_surface = old_active_surface;
 }
 
 static void
@@ -2810,7 +2866,10 @@ wl_touch_motion(void *data, struct wl_touch *wl_touch, uint32_t time,
     struct wl_window *win = wl_surface_get_user_data(seat->touch.surface);
     struct terminal *term = win->term;
 
+    struct terminal *old_term = seat->mouse_focus;
+    enum term_surface old_active_surface = term->active_surface;
     seat->mouse_focus = term;
+    term->active_surface = seat->touch.surface_kind;
 
     switch (seat->touch.state) {
     case TOUCH_STATE_HELD:
@@ -2837,7 +2896,8 @@ wl_touch_motion(void *data, struct wl_touch *wl_touch, uint32_t time,
         break;
     }
 
-    seat->mouse_focus = NULL;
+    seat->mouse_focus = old_term;
+    term->active_surface = old_active_surface;
 }
 
 static void
