@@ -41,12 +41,15 @@ enum bind_action_normal {
     BIND_ACTION_UNICODE_INPUT,
 
     /* Mouse specific actions - i.e. they require a mouse coordinate */
+    BIND_ACTION_SCROLLBACK_UP_MOUSE,
+    BIND_ACTION_SCROLLBACK_DOWN_MOUSE,
     BIND_ACTION_SELECT_BEGIN,
     BIND_ACTION_SELECT_BEGIN_BLOCK,
     BIND_ACTION_SELECT_EXTEND,
     BIND_ACTION_SELECT_EXTEND_CHAR_WISE,
     BIND_ACTION_SELECT_WORD,
     BIND_ACTION_SELECT_WORD_WS,
+    BIND_ACTION_SELECT_QUOTE,
     BIND_ACTION_SELECT_ROW,
 
     BIND_ACTION_KEY_COUNT = BIND_ACTION_UNICODE_INPUT + 1,
@@ -55,6 +58,14 @@ enum bind_action_normal {
 
 enum bind_action_search {
     BIND_ACTION_SEARCH_NONE,
+    BIND_ACTION_SEARCH_SCROLLBACK_UP_PAGE,
+    BIND_ACTION_SEARCH_SCROLLBACK_UP_HALF_PAGE,
+    BIND_ACTION_SEARCH_SCROLLBACK_UP_LINE,
+    BIND_ACTION_SEARCH_SCROLLBACK_DOWN_PAGE,
+    BIND_ACTION_SEARCH_SCROLLBACK_DOWN_HALF_PAGE,
+    BIND_ACTION_SEARCH_SCROLLBACK_DOWN_LINE,
+    BIND_ACTION_SEARCH_SCROLLBACK_HOME,
+    BIND_ACTION_SEARCH_SCROLLBACK_END,
     BIND_ACTION_SEARCH_CANCEL,
     BIND_ACTION_SEARCH_COMMIT,
     BIND_ACTION_SEARCH_FIND_PREV,
@@ -69,8 +80,14 @@ enum bind_action_search {
     BIND_ACTION_SEARCH_DELETE_PREV_WORD,
     BIND_ACTION_SEARCH_DELETE_NEXT,
     BIND_ACTION_SEARCH_DELETE_NEXT_WORD,
+    BIND_ACTION_SEARCH_EXTEND_CHAR,
     BIND_ACTION_SEARCH_EXTEND_WORD,
     BIND_ACTION_SEARCH_EXTEND_WORD_WS,
+    BIND_ACTION_SEARCH_EXTEND_LINE_DOWN,
+    BIND_ACTION_SEARCH_EXTEND_BACKWARD_CHAR,
+    BIND_ACTION_SEARCH_EXTEND_BACKWARD_WORD,
+    BIND_ACTION_SEARCH_EXTEND_BACKWARD_WORD_WS,
+    BIND_ACTION_SEARCH_EXTEND_LINE_UP,
     BIND_ACTION_SEARCH_CLIPBOARD_PASTE,
     BIND_ACTION_SEARCH_PRIMARY_PASTE,
     BIND_ACTION_SEARCH_UNICODE_INPUT,
