@@ -53,13 +53,11 @@ version_and_features(void)
 {
     static char buf[256];
     snprintf(buf, sizeof(buf),
-             "version: %s %cpgo %cime %cgraphemes %cfractional-scaling %ccursor-shape %cassertions",
+             "version: %s %cpgo %cime %cgraphemes %cassertions",
              FOOT_VERSION,
              feature_pgo() ? '+' : '-',
              feature_ime() ? '+' : '-',
              feature_graphemes() ? '+' : '-',
-             feature_fractional_scaling() ? '+' : '-',
-             feature_cursor_shape() ? '+' : '-',
              feature_assertions() ? '+' : '-');
     return buf;
 }
@@ -178,7 +176,7 @@ int
 main(int argc, char *const *argv)
 {
     /* Custom exit code, to enable users to differentiate between foot
-     * itself failing, and the client application failiing */
+     * itself failing, and the client application failing */
     static const int foot_exit_failure = -26;
     int ret = foot_exit_failure;
 
