@@ -298,6 +298,7 @@ foreach_selected(
     switch (term->selection.kind) {
     case SELECTION_CHAR_WISE:
     case SELECTION_WORD_WISE:
+    case SELECTION_QUOTE_WISE:
     case SELECTION_LINE_WISE:
         foreach_selected_normal(term, start, end, cb, data);
         return;
@@ -339,16 +340,19 @@ selection_to_text(const struct terminal *term)
     return extract_finish(ctx, &text, NULL) ? text : NULL;
 }
 
+/* Coordinates are in *absolute* row numbers (NOT view local) */
 void
-selection_find_word_boundary_left(struct terminal *term, struct coord *pos,
+selection_find_word_boundary_left(const struct terminal *term, struct coord *pos,
                                   bool spaces_only)
 {
-    xassert(pos->row >= 0);
-    xassert(pos->row < term->rows);
+    const struct grid *grid = term->grid;
+
     xassert(pos->col >= 0);
     xassert(pos->col < term->cols);
+    xassert(pos->row >= 0);
+    pos->row &= grid->num_rows - 1;
 
-    const struct row *r = grid_row_in_view(term->grid, pos->row);
+    const struct row *r = grid->rows[pos->row];
     char32_t c = r->cells[pos->col].wc;
 
     while (c >= CELL_SPACER) {
@@ -372,15 +376,22 @@ selection_find_word_boundary_left(struct terminal *term, struct coord *pos,
         int next_col = pos->col - 1;
         int next_row = pos->row;
 
-        const struct row *row = grid_row_in_view(term->grid, next_row);
+        const struct row *row = grid->rows[next_row];
 
         /* Linewrap */
         if (next_col < 0) {
             next_col = term->cols - 1;
-            if (--next_row < 0)
-                break;
 
-            row = grid_row_in_view(term->grid, next_row);
+            next_row = (next_row - 1 + grid->num_rows) & (grid->num_rows - 1);
+
+            if (grid_row_abs_to_sb(grid, term->rows, next_row) == term->grid->num_rows - 1 ||
+                grid->rows[next_row] == NULL)
+            {
+                /* Scrollback wrap-around */
+                break;
+            }
+
+            row = grid->rows[next_row];
 
             if (row->linebreak) {
                 /* Hard linebreak, treat as space. I.e. break selection */
@@ -417,17 +428,20 @@ selection_find_word_boundary_left(struct terminal *term, struct coord *pos,
     }
 }
 
+/* Coordinates are in *absolute* row numbers (NOT view local) */
 void
-selection_find_word_boundary_right(struct terminal *term, struct coord *pos,
+selection_find_word_boundary_right(const struct terminal *term, struct coord *pos,
                                    bool spaces_only,
                                    bool stop_on_space_to_word_boundary)
 {
-    xassert(pos->row >= 0);
-    xassert(pos->row < term->rows);
+    const struct grid *grid = term->grid;
+
     xassert(pos->col >= 0);
     xassert(pos->col < term->cols);
+    xassert(pos->row >= 0);
+    pos->row &= grid->num_rows - 1;
 
-    const struct row *r = grid_row_in_view(term->grid, pos->row);
+    const struct row *r = grid->rows[pos->row];
     char32_t c = r->cells[pos->col].wc;
 
     while (c >= CELL_SPACER) {
@@ -452,7 +466,7 @@ selection_find_word_boundary_right(struct terminal *term, struct coord *pos,
         int next_col = pos->col + 1;
         int next_row = pos->row;
 
-        const struct row *row = grid_row_in_view(term->grid, next_row);
+        const struct row *row = term->grid->rows[next_row];
 
         /* Linewrap */
         if (next_col >= term->cols) {
@@ -462,10 +476,14 @@ selection_find_word_boundary_right(struct terminal *term, struct coord *pos,
             }
 
             next_col = 0;
-            if (++next_row >= term->rows)
-                break;
+            next_row = (next_row + 1) & (grid->num_rows - 1);
 
-            row = grid_row_in_view(term->grid, next_row);
+            if (grid_row_abs_to_sb(grid, term->rows, next_row) == 0) {
+                /* Scrollback wrap-around */
+                break;
+            }
+
+            row = grid->rows[next_row];
         }
 
         c = row->cells[next_col].wc;
@@ -508,9 +526,86 @@ selection_find_word_boundary_right(struct terminal *term, struct coord *pos,
     }
 }
 
-void
-selection_find_line_boundary_left(struct terminal *term, struct coord *pos,
-                                  bool spaces_only)
+static bool
+selection_find_quote_left(struct terminal *term, struct coord *pos,
+                          char32_t *quote_char)
+{
+    const struct row *row = grid_row_in_view(term->grid, pos->row);
+    char32_t wc = row->cells[pos->col].wc;
+
+    if ((*quote_char == '\0' && (wc == '"' || wc == '\'')) ||
+        wc == *quote_char)
+    {
+        return false;
+    }
+
+    int next_row = pos->row;
+    int next_col = pos->col;
+
+    while (true) {
+        if (--next_col < 0) {
+            next_col = term->cols - 1;
+            if (--next_row < 0)
+                return false;
+
+            row = grid_row_in_view(term->grid, next_row);
+            if (row->linebreak)
+                return false;
+        }
+
+        wc = row->cells[next_col].wc;
+
+        if ((*quote_char == '\0' && (wc == '"' || wc == '\'')) ||
+            wc == *quote_char)
+        {
+            pos->row = next_row;
+            pos->col = next_col + 1;
+            xassert(pos->col < term->cols);
+
+            *quote_char = wc;
+            return true;
+        }
+    }
+}
+
+static bool
+selection_find_quote_right(struct terminal *term, struct coord *pos, char32_t quote_char)
+{
+    if (quote_char == '\0')
+        return false;
+
+    const struct row *row = grid_row_in_view(term->grid, pos->row);
+    char32_t wc = row->cells[pos->col].wc;
+    if (wc == quote_char)
+        return false;
+
+    int next_row = pos->row;
+    int next_col = pos->col;
+
+    while (true) {
+        if (++next_col >= term->cols) {
+            next_col = 0;
+            if (++next_row >= term->rows)
+                return false;
+
+            if (row->linebreak)
+                return false;
+
+            row = grid_row_in_view(term->grid, next_row);
+        }
+
+        wc = row->cells[next_col].wc;
+        if (wc == quote_char) {
+            pos->row = next_row;
+            pos->col = next_col - 1;
+            xassert(pos->col >= 0);
+            return true;
+        }
+    }
+}
+
+static void
+selection_find_line_boundary_left(struct terminal *term, struct coord *pos)
 {
     int next_row = pos->row;
     pos->col = 0;
@@ -530,9 +625,8 @@ selection_find_line_boundary_left(struct terminal *term, struct coord *pos,
     }
 }
 
-void
-selection_find_line_boundary_right(struct terminal *term, struct coord *pos,
-                                   bool spaces_only)
+static void
+selection_find_line_boundary_right(struct terminal *term, struct coord *pos)
 {
     int next_row = pos->row;
     pos->col = term->cols - 1;
@@ -562,6 +656,7 @@ selection_start(struct terminal *term, int col, int row,
     LOG_DBG("%s selection started at %d,%d",
             kind == SELECTION_CHAR_WISE ? "character-wise" :
             kind == SELECTION_WORD_WISE ? "word-wise" :
+            kind == SELECTION_QUOTE_WISE ? "quote-wise" :
             kind == SELECTION_LINE_WISE ? "line-wise" :
             kind == SELECTION_BLOCK ? "block" : "<unknown>",
             row, col);
@@ -581,24 +676,84 @@ selection_start(struct terminal *term, int col, int row,
         break;
 
     case SELECTION_WORD_WISE: {
-        struct coord start = {col, row}, end = {col, row};
+        struct coord start = {col, term->grid->view + row};
+        struct coord end = {col, term->grid->view + row};
         selection_find_word_boundary_left(term, &start, spaces_only);
         selection_find_word_boundary_right(term, &end, spaces_only, true);
 
-        term->selection.coords.start = (struct coord){
-            start.col, term->grid->view + start.row};
+        term->selection.coords.start = start;
 
         term->selection.pivot.start = term->selection.coords.start;
-        term->selection.pivot.end = (struct coord){end.col, term->grid->view + end.row};
+        term->selection.pivot.end = end;
 
-        selection_update(term, end.col, end.row);
+        /*
+         * FIXME: go through selection.c and make sure all public
+         * functions use the *same* coordinate system...
+         *
+         * selection_find_word_boundary*() uses absolute row numbers,
+         * while selection_update(), and pretty much all others, use
+         * view-local.
+         */
+
+        selection_update(term, end.col, end.row - term->grid->view);
         break;
     }
 
+    case SELECTION_QUOTE_WISE: {
+        struct coord start = {col, row}, end = {col, row};
+
+        char32_t quote_char = '\0';
+        bool found_left = selection_find_quote_left(term, &start, &quote_char);
+        bool found_right = selection_find_quote_right(term, &end, quote_char);
+
+        if (found_left && !found_right) {
+            xassert(quote_char != '\0');
+
+            /*
+             * Try to flip the quote character we're looking for.
+             *
+             * This lets us handle things like:
+             *
+             *   "nested 'quotes are fun', right"
+             *
+             * In the example above, starting the selection at
+             * "right", will otherwise not match. find-left will find
+             * the single quote, causing find-right to fail.
+             *
+             * By flipping the quote-character, and re-trying, we
+             * find-left will find the starting double quote, letting
+             * find-right succeed as well.
+             */
+
+            if (quote_char == '\'')
+                quote_char = '"';
+            else if (quote_char == '"')
+                quote_char = '\'';
+
+            found_left = selection_find_quote_left(term, &start, &quote_char);
+            found_right = selection_find_quote_right(term, &end, quote_char);
+        }
+
+        if (found_left && found_right) {
+            term->selection.coords.start = (struct coord){
+                start.col, term->grid->view + start.row};
+
+            term->selection.pivot.start = term->selection.coords.start;
+            term->selection.pivot.end = (struct coord){end.col, term->grid->view + end.row};
+
+            term->selection.kind = SELECTION_WORD_WISE;
+            selection_update(term, end.col, end.row);
+            break;
+        } else {
+            term->selection.kind = SELECTION_LINE_WISE;
+            /* FALLTHROUGH */
+        }
+    }
+
     case SELECTION_LINE_WISE: {
         struct coord start = {0, row}, end = {term->cols - 1, row};
-        selection_find_line_boundary_left(term, &start, spaces_only);
-        selection_find_line_boundary_right(term, &end, spaces_only);
+        selection_find_line_boundary_left(term, &start);
+        selection_find_line_boundary_right(term, &end);
 
         term->selection.coords.start = (struct coord){
             start.col, term->grid->view + start.row};
@@ -962,16 +1117,16 @@ selection_update(struct terminal *term, int col, int row)
     if (!term->selection.ongoing)
         return;
 
-    LOG_DBG("selection updated: start = %d,%d, end = %d,%d -> %d, %d",
-            term->selection.coords.start.row, term->selection.coords.start.col,
-            term->selection.coords.end.row, term->selection.coords.end.col,
-            row, col);
-
     xassert(term->grid->view + row != -1);
 
     struct coord new_start = term->selection.coords.start;
     struct coord new_end = {col, term->grid->view + row};
 
+    LOG_DBG("selection updated: start = %d,%d, end = %d,%d -> %d, %d",
+            term->selection.coords.start.row, term->selection.coords.start.col,
+            term->selection.coords.end.row, term->selection.coords.end.col,
+            new_end.row, new_end.col);
+
     /* Adjust start point if the selection has changed 'direction' */
     if (!(new_end.row == new_start.row && new_end.col == new_start.col)) {
         enum selection_direction new_direction = term->selection.direction;
@@ -1031,41 +1186,39 @@ selection_update(struct terminal *term, int col, int row)
 
     case SELECTION_WORD_WISE:
         switch (term->selection.direction) {
-        case SELECTION_LEFT: {
-            struct coord end = {col, row};
+        case SELECTION_LEFT:
+            new_end = (struct coord){col, term->grid->view + row};
             selection_find_word_boundary_left(
-                term, &end, term->selection.spaces_only);
-            new_end = (struct coord){end.col, term->grid->view + end.row};
+                term, &new_end, term->selection.spaces_only);
             break;
-        }
 
-        case SELECTION_RIGHT: {
-            struct coord end = {col, row};
+        case SELECTION_RIGHT:
+            new_end = (struct coord){col, term->grid->view + row};
             selection_find_word_boundary_right(
-                term, &end, term->selection.spaces_only, true);
-            new_end = (struct coord){end.col, term->grid->view + end.row};
+                term, &new_end, term->selection.spaces_only, true);
             break;
-        }
 
         case SELECTION_UNDIR:
             break;
         }
         break;
 
+    case SELECTION_QUOTE_WISE:
+        BUG("quote-wise selection should always be transformed to either word-wise or line-wise");
+        break;
+
     case SELECTION_LINE_WISE:
         switch (term->selection.direction) {
         case SELECTION_LEFT: {
             struct coord end = {0, row};
-            selection_find_line_boundary_left(
-                term, &end, term->selection.spaces_only);
+            selection_find_line_boundary_left(term, &end);
             new_end = (struct coord){end.col, term->grid->view + end.row};
             break;
         }
 
         case SELECTION_RIGHT: {
             struct coord end = {col, row};
-            selection_find_line_boundary_right(
-                term, &end, term->selection.spaces_only);
+            selection_find_line_boundary_right(term, &end);
             new_end = (struct coord){end.col, term->grid->view + end.row};
             break;
         }
@@ -1215,16 +1368,19 @@ selection_extend_normal(struct terminal *term, int col, int row,
         xassert(new_kind == SELECTION_CHAR_WISE ||
                new_kind == SELECTION_WORD_WISE);
 
-        struct coord pivot_start = {new_start.col, new_start.row - term->grid->view};
+        struct coord pivot_start = {new_start.col, new_start.row};
         struct coord pivot_end = pivot_start;
 
         selection_find_word_boundary_left(term, &pivot_start, spaces_only);
         selection_find_word_boundary_right(term, &pivot_end, spaces_only, true);
 
-        term->selection.pivot.start =
-            (struct coord){pivot_start.col, term->grid->view + pivot_start.row};
-        term->selection.pivot.end =
-            (struct coord){pivot_end.col, term->grid->view + pivot_end.row};
+        term->selection.pivot.start = pivot_start;
+        term->selection.pivot.end = pivot_end;
+        break;
+    }
+
+    case SELECTION_QUOTE_WISE: {
+        BUG("quote-wise selection should always be transformed to either word-wise or line-wise");
         break;
     }
 
@@ -1235,8 +1391,8 @@ selection_extend_normal(struct terminal *term, int col, int row,
         struct coord pivot_start = {new_start.col, new_start.row - term->grid->view};
         struct coord pivot_end = pivot_start;
 
-        selection_find_line_boundary_left(term, &pivot_start, spaces_only);
-        selection_find_line_boundary_right(term, &pivot_end, spaces_only);
+        selection_find_line_boundary_left(term, &pivot_start);
+        selection_find_line_boundary_right(term, &pivot_end);
 
         term->selection.pivot.start =
             (struct coord){pivot_start.col, term->grid->view + pivot_start.row};
@@ -1362,6 +1518,7 @@ selection_extend(struct seat *seat, struct terminal *term,
 
     case SELECTION_CHAR_WISE:
     case SELECTION_WORD_WISE:
+    case SELECTION_QUOTE_WISE:
     case SELECTION_LINE_WISE:
         selection_extend_normal(term, col, row, new_kind);
         break;
