@@ -5,7 +5,8 @@
 #include "reaper.h"
 
 bool spawn(struct reaper *reaper, const char *cwd, char *const argv[],
-           int stdin_fd, int stdout_fd, int stderr_fd);
+           int stdin_fd, int stdout_fd, int stderr_fd,
+           const char *xdg_activation_token);
 
 bool spawn_expand_template(
     const struct config_spawn_template *template,
