@@ -86,7 +86,6 @@ spawn_url_launcher_with_token(struct terminal *term,
     return ret;
 }
 
-#if defined(HAVE_XDG_ACTIVATION)
 struct spawn_activation_context {
     struct terminal *term;
     char *url;
@@ -101,13 +100,11 @@ activation_token_done(const char *token, void *data)
     free(ctx->url);
     free(ctx);
 }
-#endif
 
 static bool
 spawn_url_launcher(struct seat *seat, struct terminal *term, const char *url,
                    uint32_t serial)
 {
-#if defined(HAVE_XDG_ACTIVATION)
     struct spawn_activation_context *ctx = xmalloc(sizeof(*ctx));
     *ctx = (struct spawn_activation_context){
         .term = term,
@@ -123,7 +120,6 @@ spawn_url_launcher(struct seat *seat, struct terminal *term, const char *url,
 
     free(ctx->url);
     free(ctx);
-#endif
 
     return spawn_url_launcher_with_token(term, url, NULL);
 }
@@ -132,46 +128,17 @@ static void
 activate_url(struct seat *seat, struct terminal *term, const struct url *url,
              uint32_t serial)
 {
-    char *url_string = NULL;
-
-    char *scheme, *host, *path;
-    if (uri_parse(url->url, strlen(url->url), &scheme, NULL, NULL,
-                  &host, NULL, &path, NULL, NULL))
-    {
-        if (strcmp(scheme, "file") == 0 && hostname_is_localhost(host)) {
-            /*
-             * This is a file in *this* computer. Pass only the
-             * filename to the URL-launcher.
-             *
-             * I.e. strip the ‘file://user@host/’ prefix.
-             */
-            url_string = path;
-        } else
-            free(path);
-
-        free(scheme);
-        free(host);
-    }
-
-    if (url_string == NULL)
-        url_string = xstrdup(url->url);
-
     switch (url->action) {
     case URL_ACTION_COPY:
-        if (text_to_clipboard(seat, term, url_string, seat->kbd.serial)) {
-            /* Now owned by our clipboard “manager” */
-            url_string = NULL;
-        }
+        text_to_clipboard(seat, term, xstrdup(url->url), seat->kbd.serial);
         break;
 
     case URL_ACTION_LAUNCH:
     case URL_ACTION_PERSISTENT: {
-        spawn_url_launcher(seat, term, url_string, serial);
+        spawn_url_launcher(seat, term, url->url, serial);
         break;
     }
     }
-
-    free(url_string);
 }
 
 void
@@ -330,121 +297,138 @@ auto_detected(const struct terminal *term, enum url_action action,
 
         for (int c = 0; c < term->cols; c++) {
             const struct cell *cell = &row->cells[c];
-            char32_t wc = cell->wc;
 
-            switch (state) {
-            case STATE_PROTOCOL:
-                for (size_t i = 0; i < max_prot_len - 1; i++) {
+            if (cell->wc >= CELL_SPACER)
+                continue;
+
+            const char32_t *wcs = NULL;
+            size_t wc_count = 0;
+
+            if (cell->wc >= CELL_COMB_CHARS_LO && cell->wc <= CELL_COMB_CHARS_HI) {
+                struct composed *composed =
+                    composed_lookup(term->composed, cell->wc - CELL_COMB_CHARS_LO);
+                wcs = composed->chars;
+                wc_count = composed->count;
+            } else {
+                wcs = &cell->wc;
+                wc_count = 1;
+            }
+
+            for (size_t w_idx = 0; w_idx < wc_count; w_idx++) {
+                char32_t wc = wcs[w_idx];
+
+                switch (state) {
+                case STATE_PROTOCOL:
+                  for (size_t i = 0; i < max_prot_len - 1; i++) {
                     proto_chars[i] = proto_chars[i + 1];
                     proto_start[i] = proto_start[i + 1];
-                }
+                  }
 
-                if (proto_char_count >= max_prot_len)
+                  if (proto_char_count >= max_prot_len)
                     proto_char_count = max_prot_len - 1;
 
-                proto_chars[max_prot_len - 1] = wc;
-                proto_start[max_prot_len - 1] = (struct coord){c, r};
-                proto_char_count++;
+                  proto_chars[max_prot_len - 1] = wc;
+                  proto_start[max_prot_len - 1] = (struct coord){c, r};
+                  proto_char_count++;
 
-                for (size_t i = 0; i < conf->url.prot_count; i++) {
+                  for (size_t i = 0; i < conf->url.prot_count; i++) {
                     size_t prot_len = c32len(conf->url.protocols[i]);
 
                     if (proto_char_count < prot_len)
-                        continue;
+                      continue;
 
-                    const char32_t *proto = &proto_chars[max_prot_len - prot_len];
+                    const char32_t *proto =
+                        &proto_chars[max_prot_len - prot_len];
 
-                    if (c32ncasecmp(conf->url.protocols[i], proto, prot_len) == 0) {
-                        state = STATE_URL;
-                        start = proto_start[max_prot_len - prot_len];
+                    if (c32ncasecmp(conf->url.protocols[i], proto, prot_len) ==
+                        0) {
+                      state = STATE_URL;
+                      start = proto_start[max_prot_len - prot_len];
 
-                        c32ncpy(url, proto, prot_len);
-                        len = prot_len;
+                      c32ncpy(url, proto, prot_len);
+                      len = prot_len;
 
-                        parenthesis = brackets = ltgts = 0;
-                        break;
+                      parenthesis = brackets = ltgts = 0;
+                      break;
                     }
-                }
-                break;
+                  }
+                  break;
 
-            case STATE_URL: {
-                const char32_t *match = bsearch(
-                    &wc,
-                    uri_characters,
-                    uri_characters_count,
-                    sizeof(uri_characters[0]),
-                    &c32cmp_single);
+                case STATE_URL: {
+                  const char32_t *match =
+                      bsearch(&wc, uri_characters, uri_characters_count,
+                              sizeof(uri_characters[0]), &c32cmp_single);
 
-                bool emit_url = false;
+                  bool emit_url = false;
 
-                if (match == NULL) {
+                  if (match == NULL) {
                     /*
                      * Character is not a valid URI character. Emit
                      * the URL we’ve collected so far, *without*
                      * including _this_ character.
                      */
                     emit_url = true;
-                } else {
+                  } else {
                     xassert(*match == wc);
 
                     switch (wc) {
                     default:
-                        url[len++] = wc;
-                        break;
+                      url[len++] = wc;
+                      break;
 
                     case U'(':
-                        parenthesis++;
-                        url[len++] = wc;
-                        break;
+                      parenthesis++;
+                      url[len++] = wc;
+                      break;
 
                     case U'[':
-                        brackets++;
-                        url[len++] = wc;
-                        break;
+                      brackets++;
+                      url[len++] = wc;
+                      break;
 
                     case U'<':
-                        ltgts++;
-                        url[len++] = wc;
-                        break;
+                      ltgts++;
+                      url[len++] = wc;
+                      break;
 
                     case U')':
-                        if (--parenthesis < 0)
-                            emit_url = true;
-                        else
-                            url[len++] = wc;
-                        break;
+                      if (--parenthesis < 0)
+                        emit_url = true;
+                      else
+                        url[len++] = wc;
+                      break;
 
                     case U']':
-                        if (--brackets < 0)
-                            emit_url = true;
-                        else
-                            url[len++] = wc;
-                        break;
+                      if (--brackets < 0)
+                        emit_url = true;
+                      else
+                        url[len++] = wc;
+                      break;
 
                     case U'>':
-                        if (--ltgts < 0)
-                            emit_url = true;
-                        else
-                            url[len++] = wc;
-                        break;
+                      if (--ltgts < 0)
+                        emit_url = true;
+                      else
+                        url[len++] = wc;
+                      break;
                     }
-                }
+                  }
 
-                if (c >= term->cols - 1 && row->linebreak) {
+                  if (c >= term->cols - 1 && row->linebreak) {
                     /*
                      * Endpoint is inclusive, and we’ll be subtracting
                      * 1 from the column when emitting the URL.
                      */
                     c++;
                     emit_url = true;
-                }
+                  }
 
-                if (emit_url) {
+                  if (emit_url) {
                     struct coord end = {c, r};
 
                     if (--end.col < 0) {
-                        end.row--;
-                        end.col = term->cols - 1;
+                      end.row--;
+                      end.col = term->cols - 1;
                     }
 
                     /* Heuristic to remove trailing characters that
@@ -452,21 +436,28 @@ auto_detected(const struct terminal *term, enum url_action action,
                      * the end of the URL */
                     bool done = false;
                     do {
-                        switch (url[len - 1]) {
-                        case U'.': case U',': case U':': case U';': case U'?':
-                        case U'!': case U'"': case U'\'': case U'%':
-                            len--;
-                            end.col--;
-                            if (end.col < 0) {
-                                end.row--;
-                                end.col = term->cols - 1;
-                            }
-                            break;
-
-                        default:
-                            done = true;
-                            break;
+                      switch (url[len - 1]) {
+                      case U'.':
+                      case U',':
+                      case U':':
+                      case U';':
+                      case U'?':
+                      case U'!':
+                      case U'"':
+                      case U'\'':
+                      case U'%':
+                        len--;
+                        end.col--;
+                        if (end.col < 0) {
+                          end.row--;
+                          end.col = term->cols - 1;
                         }
+                        break;
+
+                      default:
+                        done = true;
+                        break;
+                      }
                     } while (!done);
 
                     url[len] = U'\0';
@@ -476,25 +467,26 @@ auto_detected(const struct terminal *term, enum url_action action,
 
                     char *url_utf8 = ac32tombs(url);
                     if (url_utf8 != NULL) {
-                        tll_push_back(
-                            *urls,
-                            ((struct url){
-                                .id = (uint64_t)rand() << 32 | rand(),
-                                .url = url_utf8,
-                                .range = {
-                                    .start = start,
-                                    .end = end,
-                                },
-                                .action = action,
-                                .osc8 = false}));
+                      tll_push_back(
+                          *urls,
+                          ((struct url){.id = (uint64_t)rand() << 32 | rand(),
+                                        .url = url_utf8,
+                                        .range =
+                                            {
+                                                .start = start,
+                                                .end = end,
+                                            },
+                                        .action = action,
+                                        .osc8 = false}));
                     }
 
                     state = STATE_PROTOCOL;
                     len = 0;
                     parenthesis = brackets = ltgts = 0;
+                  }
+                  break;
+                }
                 }
-                break;
-            }
             }
         }
     }
@@ -705,7 +697,7 @@ urls_assign_key_combos(const struct config *conf, url_list_t *urls)
 
         /*
          * Scan previous URLs, and check if *this* URL matches any of
-         * them; if so, re-use the *same* key combo.
+         * them; if so, reuse the *same* key combo.
          */
         bool url_already_seen = false;
         tll_foreach(*urls, it2) {
